from enum import Enum


class DataType(Enum):
    # these need to match the ones in nema-artifacts-model
    STRING = "STRING"
    INT = "INT"
    FLOAT = "FLOAT"
    BOOL = "BOOL"
    CURRENCY = "CURRENCY"
    PERCENTAGE = "PERCENTAGE"
    INT_WITH_PHYSICAL_UNIT_V0 = "INT_WITH_PHYSICAL_UNIT.V0"
    FLOAT_WITH_PHYSICAL_UNIT_V0 = "FLOAT_WITH_PHYSICAL_UNIT.V0"
    INT_WITH_ARBITRARY_UNIT_V0 = "INT_WITH_ARBITRARY_UNIT.V0"
    FLOAT_WITH_ARBITRARY_UNIT_V0 = "FLOAT_WITH_ARBITRARY_UNIT.V0"
    LAT_LONG_EARTH_COORDINATE_V0 = "LAT_LONG_EARTH_COORDINATE.V0"
    LINE_FIGURE_V0 = "LINE_FIGURE.V0"
    SUNBURST_FIGURE_V0 = "SUNBURST_FIGURE.V0"
    VEGA_FIGURE_V0 = "VEGA_FIGURE.V0"
    CSV_V0 = "CSV.V0"
    IMAGE_V0 = "IMAGE.V0"
    ARBITRARY_FILE_V0 = "ARBITRARY_FILE.V0"
    ARBITRARY_FILE_COLLECTION_V0 = "ARBITRARY_FILE_COLLECTION.V0"
    DATE_V0 = "DATE.V0"
    DATE_TIME_V0 = "DATE_TIME.V0"



import pandas as pd
from dataclasses import dataclass
from typing import Optional
import os

from nema.utils.file_name import generate_random_file_name
from nema.data.data_properties import BlobDataProperties


@dataclass
class TableDataProperties(BlobDataProperties):
    pass


@dataclass
class CSVData(TableDataProperties):
    df: Optional[pd.DataFrame] = None

    def __nema_marshall__(self):
        return {}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls()

    def get_value(self):
        return self.df

    def write_data_to_file_and_return_file_name(self, destination_folder: str):
        # move to destination folder
        output_file_name = generate_random_file_name("csv")
        destination_file_path = os.path.join(destination_folder, output_file_name)

        self.df.to_csv(destination_file_path, index=False)

        return output_file_name

    def get_contents(self) -> pd.DataFrame:
        return self.df

    def extract_contents(self, file_path: str):
        self.df = pd.read_csv(file_path)

        return self.df

    @property
    def data_type(self):
        return "CSV.V0"



from nema.data.data_type import DataType
from nema.data.data_properties import (
    StringValue,
    IntegerValue,
    FloatValue,
    CurrencyValue,
    FloatValueWithArbitraryUnit,
    IntValueWithArbitraryUnit,
    FloatValueWithPhysicalUnit,
    IntValueWithPhysicalUnit,
)


def map_type_to_data_properties(data_type: DataType):
    if data_type == DataType.STRING:
        return StringValue
    elif data_type == DataType.INT:
        return IntegerValue
    elif data_type == DataType.FLOAT:
        return FloatValue
    elif data_type == DataType.CURRENCY:
        return CurrencyValue
    elif data_type == DataType.FLOAT_WITH_ARBITRARY_UNIT_V0:
        return FloatValueWithArbitraryUnit
    elif data_type == DataType.INT_WITH_ARBITRARY_UNIT_V0:
        return IntValueWithArbitraryUnit
    elif data_type == DataType.FLOAT_WITH_PHYSICAL_UNIT_V0:
        return FloatValueWithPhysicalUnit
    elif data_type == DataType.INT_WITH_PHYSICAL_UNIT_V0:
        return IntValueWithPhysicalUnit
    else:
        raise ValueError(f"Data type {data_type} not supported.")



import toml
import os
from dataclasses import dataclass, field
from typing import Dict, List

from .url import get_ids_from_app_url

CONFIG_FILE_PATH = "nema.toml"

# and remove the trailing slash
API_URL_RAW = os.environ.get("NEMA_API_URL", "https://api.nemasystems.io")
API_URL = API_URL_RAW.rstrip("/")


@dataclass
class GlobalConfigWorkflow:
    global_id: int = 0
    key: str = ""
    name: str = ""
    description: str = ""
    command: List[str] = field(default_factory=list)

    def _marshall(self):
        return {
            "global_id": self.global_id,
            "name": self.name,
            "description": self.description,
            "command": self.command,
        }


@dataclass
class GlobalConfig:

    _is_set: bool = False

    project_url: str = ""

    _project_id: str = ""
    _workspace_id: str = ""
    _tenant_id: str = ""

    workflows: Dict[str, GlobalConfigWorkflow] = field(default_factory=dict)

    def __post_init__(self):

        if os.path.exists(CONFIG_FILE_PATH):
            config = toml.load(CONFIG_FILE_PATH)

            self.project_url = config.get("project_url")
            self._is_set = True

            ids = get_ids_from_app_url(self.project_url)
            self._tenant_id = ids.tenant_id
            self._workspace_id = ids.workspace_id
            self._project_id = ids.project_id

            # record workflows
            workflows_config = config.get("workflows", {})
            new_workflows = {}
            for key, workflow_config in workflows_config.items():
                new_workflow = GlobalConfigWorkflow(key=key, **workflow_config)
                new_workflows[key] = new_workflow
            self.workflows = new_workflows

        else:
            self._is_set = False

    @property
    def is_set(self) -> bool:
        return self._is_set

    @property
    def tenant_id(self) -> bool:
        return self._tenant_id

    @property
    def workspace_id(self) -> bool:
        return self._workspace_id

    @property
    def project_id(self) -> bool:
        return self._project_id

    @property
    def tenant_api_url(self) -> str:
        return f"{API_URL}/app/{self.tenant_id}"

    @property
    def project_api_url(self) -> str:
        return f"{self.tenant_api_url}/{self.workspace_id}/{self.project_id}"

    def _marshall(self):
        return {
            "project_url": self.project_url,
            "workflows": {
                key: workflow._marshall() for key, workflow in self.workflows.items()
            },
        }

    def save(self):
        d = self._marshall()
        with open(CONFIG_FILE_PATH, "w") as f:
            f.write(toml.dumps(d))

    def update_configs(self, project_id: str, workspace_id: str, tenant_id: str):
        self._project_id = project_id
        self._workspace_id = workspace_id
        self._tenant_id = tenant_id
        self._is_set = True


GLOBAL_CONFIG = GlobalConfig()



from typing import Callable
from functools import wraps
import uuid


from nema.data.data_collection import DataCollection
from nema.workflow import Workflow


def run(
    data_collection: DataCollection,
    enable_API_calls=True,
    job_id=None,
):
    # create random string if not provided
    actual_job_id = job_id if job_id else str(uuid.uuid4())

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):

            if enable_API_calls:
                data_collection.sync_data_from_API(actual_job_id)

            result = func(*args, **kwargs)

            data_collection.save_updates_locally(actual_job_id)

            return result

        return wrapper

    # close all the data properties in the collection
    data_collection.close()

    return decorator



from dataclasses import dataclass, field
from datetime import datetime
from typing import Type, IO, Optional
from copy import deepcopy
import json

from nema.connectivity import ConnectivityManager

from .data_properties import DataProperties
from .map_type_to_data_properties import map_type_to_data_properties
from .data_type import DataType


@dataclass
class CacheInfo:
    last_time_updated: datetime = datetime(1970, 1, 1)
    number_of_times_hit: int = 0
    in_sync_cloud: bool = False


@dataclass
class Data:
    _global_id: str
    _data: DataProperties
    _cache_info: CacheInfo = field(default_factory=CacheInfo)
    _userdefined_id: Optional[str] = None
    _commit_id: Optional[str] = None
    _branch: Optional[str] = None

    @classmethod
    def init_from_cloud(
        cls,
        global_id: str,
        data_class: Type[DataProperties],
        userdefined_id: Optional[str] = None,
        commit_id: Optional[str] = None,
        **kwargs,
    ):
        data = data_class(**kwargs)

        return cls(
            _global_id=global_id,
            _userdefined_id=userdefined_id,
            _data=data,
            _commit_id=commit_id,
        )

    @classmethod
    def init_from_cloud_and_download(
        cls,
        global_id: str,
        commit_id: Optional[str] = None,
        branch: Optional[str] = None,
    ):
        connectivity_manager = ConnectivityManager()
        raw_data = connectivity_manager.pull_in_memory_data(
            global_id,
            commit_id=commit_id,
            branch=branch,
        )

        data_type = DataType(raw_data["data_type"])

        data_properties_class = map_type_to_data_properties(data_type)

        data = data_properties_class.__nema_unmarshall__(raw_data["data_properties"])

        return cls(
            _global_id=global_id,
            _data=data,
            _commit_id=commit_id,
            _branch=branch,
        )

    @classmethod
    def init_from_properties(
        cls,
        global_id: str,
        data_properties: DataProperties,
        userdefined_id: Optional[str] = None,
        commit_id: Optional[str] = None,
    ):
        return cls(
            _global_id=global_id,
            _data=data_properties,
            _userdefined_id=userdefined_id,
            _commit_id=commit_id,
        )

    @property
    def global_id(self):
        return self._global_id

    @property
    def data(self):
        # update cache info
        self._cache_info.number_of_times_hit += 1

        return self._data

    @property
    def value(self):
        return self.data.get_value()

    def update_data(self, new_data: DataProperties):
        if self._cache_info.number_of_times_hit > 0:
            raise ValueError("Data can only be updated if it has not been accessed")

        self._data = new_data
        self._cache_info.last_time_updated = datetime.now()
        self._cache_info.number_of_times_hit = 0
        self._cache_info.in_sync_cloud = False

    @property
    def userdefined_id(self):
        return self._userdefined_id

    @userdefined_id.setter
    def userdefined_id(self, value: str):
        self._userdefined_id = value

    @property
    def cache_info(self):
        return deepcopy(self._cache_info)

    def sync_from_API(
        self,
        connectivity_manager: ConnectivityManager,
        input_folder: str = "",
        commit_id: Optional[str] = None,
    ):
        # Sync data from the API

        raw_data = connectivity_manager.pull_in_memory_data(
            self.global_id, commit_id=commit_id
        )

        # we need to use _data here, otherwise we record a cache hit
        self._data = self._data.__nema_unmarshall__(raw_data["data_properties"])

    def marshall_data_properties(self):
        return self._data.__nema_marshall__()

    @property
    def is_updated(self):
        return (
            self._cache_info.in_sync_cloud == False
            and self._cache_info.last_time_updated.year > 1980
        )

    @property
    def data_type(self):
        return self._data.data_type

    def process_output_data(self, destination_folder: str):

        core_dict = {
            "global_id": self.global_id,
            "data_properties": self.marshall_data_properties(),
            "data_type": self.data_type,
        }

        if self._data.is_blob_data:
            core_dict["file_name"] = self._data.write_data_to_file_and_return_file_name(
                destination_folder=destination_folder
            )

        return core_dict

    def close(self):
        return self._data.close()


@dataclass
class FileData(Data):

    _input_folder: Optional[str] = None
    _file_reference: Optional[IO] = None

    def sync_from_API(self, *args, input_folder: str = "", **kwargs):
        super().sync_from_API(*args, input_folder=input_folder, **kwargs)
        self._input_folder = input_folder

    def get_file_path(self, requested_filename: Optional[str] = None):
        file_name = self.data.get_file_name(requested_filename)

        if file_name is None or file_name == "":
            local_file_name = self.download_blob_data(requested_filename)
            self._data.set_file_name(local_file_name, requested_filename)
            return f"{self._input_folder}/{local_file_name}"
        else:
            return f"{self._input_folder}/{file_name}"

    @property
    def contents(self):
        contents = (
            self.data.get_contents()
        )  # only use self.data once in this function to ensure cache hit is only recorded once

        if contents is None:
            # need to download the file
            filename = self.download_blob_data()
            return self._data.extract_contents(
                file_path=f"{self._input_folder}/{filename}"
            )
        else:
            return contents

    def download_blob_data(self, requested_filename: Optional[str] = None):
        connectivity_manager = ConnectivityManager()

        return connectivity_manager.pull_blob_data(
            self.global_id,
            folder=self._input_folder,
            commit_id=self._commit_id,
            requested_filename=requested_filename,
            branch=self._branch,
        )

    def get_file_name_to_save(self):
        self._cache_info.last_time_updated = datetime.now()
        self._cache_info.number_of_times_hit = 0
        self._cache_info.in_sync_cloud = False

        return self._data.get_file_name_to_save()

    # This method is called when entering the 'with' block, and it accepts the mode
    def __enter__(self):
        # Dynamically set the mode (use 'r' by default if not provided)
        self._file_reference = open(self.file_path, self._file_reading_mode)
        return self._file_reference

    # This method is called when exiting the 'with' block
    def __exit__(self, exc_type, exc_value, traceback):
        if self._file_reference:
            self._file_reference.close()  # Ensure the file is closed properly

    # This will allow 'with open()' to pass the mode and other arguments
    def __call__(self, file_reading_mode: str):
        self._file_reading_mode = file_reading_mode
        return self



from dataclasses import dataclass
from typing import Optional, List
import json
import os

from nema.connectivity import ConnectivityManager
from nema.data import Data


@dataclass
class Workflow:
    global_id: int
    name: str = ""
    description: str = ""
    output_folder: Optional[str] = None

    def marshall_create(self):
        return {
            "name": self.name,
            "description": self.description,
            "workflow_type": "NEMA.EXTERNAL_PYTHON.V0",
            "workflow_properties": {},
        }

    def create(self):
        conn = ConnectivityManager()
        global_id = conn.create_workflow(self.marshall_create())
        return global_id

    def process_update(self):
        conn = ConnectivityManager()

        # read data from output folder
        if self.output_folder:
            with open(os.path.join(self.output_folder, "output.json"), "r") as f:
                raw_data = json.load(f)
        else:
            raise ValueError("Output folder not provided")

        # read files from output folder
        files = []
        files_output_folder = os.path.join(self.output_folder, "data-output")
        for file_name in os.listdir(files_output_folder):
            files.append(os.path.join(files_output_folder, file_name))

        conn.push_workflow_update(self.global_id, raw_data, files)

    def push_outputs_to_API(
        self,
        updated_data: List[Data],
        branch_name: Optional[str] = None,
    ):
        conn = ConnectivityManager()

        output_folder = self.output_folder or ".nema"

        files_output_folder = os.path.join(output_folder, "data-output")

        os.makedirs(files_output_folder, exist_ok=True)

        raw_data = [
            (
                "updated_data",
                json.dumps(
                    data.process_output_data(destination_folder=files_output_folder)
                ),
            )
            for data in updated_data
        ]

        # read files from output folder
        files = []
        for file_name in os.listdir(files_output_folder):
            files.append(os.path.join(files_output_folder, file_name))

        conn.push_workflow_outputs(
            self.global_id, raw_data, branch=branch_name, files=files
        )



from dataclasses import dataclass


@dataclass
class IDs:
    tenant_id: str
    workspace_id: str
    project_id: str


def get_ids_from_app_url(app_url: str) -> IDs:

    app_url_without_https = app_url.replace("https://", "").replace("http://", "")

    # check if the project is pooled
    if "https://app.nemasystems.io" in app_url or "http://localhost" in app_url:
        url_splits = app_url_without_https.split("/")
        tenant_id = url_splits[1]
        workspace_id = url_splits[2]
        project_id = url_splits[3]
    else:
        # tenant is siloed
        tenant_id = app_url_without_https.split(".")[0]
        url_splits = app_url_without_https.split("/")
        workspace_id = url_splits[1]
        project_id = url_splits[2]

    return IDs(
        tenant_id=tenant_id,
        workspace_id=workspace_id,
        project_id=project_id,
    )



from dataclasses import dataclass
from typing import Dict, Optional
import json
import os

from nema.connectivity import ConnectivityManager

from .data import Data


@dataclass
class DataCollection:
    items: Dict[str, Data]

    def __getattr__(self, item_id: str):
        try:
            return self.items[item_id]
        except KeyError:
            raise AttributeError(
                f"This DataCollection object has no attribute '{item_id}'"
            )

    def sync_data_from_API(self, job_id: Optional[str] = None):
        # Sync data from the API

        # Retrieve commit ID to ensure that all the data uses the same commit ID
        conn_manager = ConnectivityManager()
        latest_commit_id = conn_manager.get_latest_commit()

        input_fldr = f".nema/input/{job_id}"

        os.makedirs(input_fldr, exist_ok=True)

        for item in self.items.values():
            item.sync_from_API(
                connectivity_manager=conn_manager,
                input_folder=input_fldr,
                commit_id=latest_commit_id,
            )

    def save_updates_locally(self, job_id: str):
        # Process updates

        # find which items have been used
        filtered_input_data = filter(
            lambda x: x.cache_info.number_of_times_hit > 0,
            self.items.values(),
        )

        # find which items have been updated
        filtered_output_data = filter(lambda x: x.is_updated, self.items.values())

        # output data to folder
        folder = os.path.join(".nema/output", job_id)
        os.makedirs(folder, exist_ok=True)

        data_output_folder = os.path.join(folder, "data-output")
        os.makedirs(data_output_folder, exist_ok=True)

        raw_data = [
            *[("used_data", data.global_id) for data in filtered_input_data],
            *[
                (
                    "updated_data",
                    json.dumps(
                        data.process_output_data(destination_folder=data_output_folder)
                    ),
                )
                for data in filtered_output_data
            ],
        ]

        output_file = os.path.join(folder, "output.json")
        with open(output_file, "w") as f:
            f.write(json.dumps(raw_data))

        # write to terminal the output file location
        print(f"Nema output written to folder: {folder}")

    def close(self):
        for item in self.items.values():
            item.close()



from dataclasses import dataclass
from typing import Optional

from nema.connectivity import ConnectivityManager


@dataclass
class App:
    global_id: int
    name: str = ""
    description: str = ""
    output_folder: Optional[str] = None

    def download_code(self):
        conn = ConnectivityManager()

        return conn.retrieve_app_code(self.global_id)



import click
import os
import subprocess
import requests
from typing import Optional
import re

from nema.utils.global_config import GLOBAL_CONFIG, GlobalConfig, GlobalConfigWorkflow
from nema.workflow.workflow import Workflow
from nema.connectivity.connectivity_manager import save_auth_data

NEMA_PATH = os.path.expanduser("~/.nema")
AUTH_PATH = os.path.join(NEMA_PATH, "auth")


def authenticate_user(username, password):

    # Authenticate the user
    if GLOBAL_CONFIG.is_set:

        tenant_url = GLOBAL_CONFIG.tenant_api_url

        login_url = f"{tenant_url}/authentication/login"

    else:
        raise Exception(
            "There is no nema.toml file in this directory. Please run `nema init` to create one."
        )

    # Make a request to the login URL
    response = requests.post(
        login_url, json={"username": username, "password": password}
    )

    if not response.ok:
        if response.status_code == 401:
            print("Invalid credentials")
        else:
            print(response.status_code, response.text)
            print("Failed to login")
        return None

    # save the authentication and refresh token to the home directory
    response_data = response.json()
    tokens = response_data["tokens"]
    refresh_token = tokens["refresh_token"]
    access_token = tokens["access_token"]

    # create the directory if it does not exist
    save_auth_data(refresh_token=refresh_token, access_token=access_token)

    print("Login successful \U00002705")


@click.group()
def cli():
    pass


@cli.command()
def login():
    """Login to Nema."""
    username = click.prompt("Please enter your username or email address")
    password = click.prompt("Please enter your password", hide_input=True)
    authenticate_user(username, password)


@cli.command()
def init():
    "Initialize nema.toml file"
    print("Initializing nema.toml file")

    new_global_config = GlobalConfig()

    new_global_config.project_url = click.prompt(
        "Please enter the project URL", type=str
    )

    new_workflow_key = click.prompt(
        "Please enter a workflow identifier (this is only used locally)",
        type=str,
        default="my-first-workflow",
    )
    new_workflow_name = click.prompt(
        "Please enter a name for the workflow", type=str, default=new_workflow_key
    )
    new_workflow_description = click.prompt(
        "Please enter a workflow description",
        type=str,
        default="A Python workflow",
    )

    new_workflow = GlobalConfigWorkflow(
        key=new_workflow_key,
        name=new_workflow_name,
        description=new_workflow_description,
        command=["python", "main.py"],
    )

    new_global_config.workflows[new_workflow_key] = new_workflow

    new_global_config.save()


@cli.group()
def workflow():
    pass


@workflow.command()
@click.argument("identifier", required=False)
def init(identifier: Optional[str]):
    "Create a new workflow"

    if identifier is None:
        all_identifiers = GLOBAL_CONFIG.workflows.keys()
    else:
        print("Initializing all workflows")
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Initializing workflow with identifier '{this_identifier}'")
        existing_workflow = GLOBAL_CONFIG.workflows[this_identifier]

        if existing_workflow.global_id > 0:
            print("Workflow already exists. Skipping.")
            continue

        workflow = Workflow(
            global_id=0,
            name=existing_workflow.name,
            description=existing_workflow.description,
        )

        global_id = workflow.create()
        print(f"Workflow successfully created with global id {global_id}. \U00002705")

        existing_workflow.global_id = global_id

    # save the global ID to the config file
    GLOBAL_CONFIG.save()


@workflow.command()
@click.argument("identifier", required=False)
def run(identifier: Optional[str]):
    "Run the workflow"
    if identifier is None:
        all_identifiers = GLOBAL_CONFIG.workflows.keys()
    else:
        print("Initializing all workflows")
        all_identifiers = [identifier]

    for this_identifier in all_identifiers:
        print(f"Running workflow with identifier '{this_identifier}'")

        existing_workflow = GLOBAL_CONFIG.workflows[this_identifier]

        # run the python file
        subprocess_result = subprocess.run(
            existing_workflow.command, capture_output=True, text=True
        )
        print(subprocess_result.stdout)
        if subprocess_result.returncode == 0:
            output = subprocess_result.stdout

            match = re.search(r"Nema output written to folder: (.+)", output)
            if match:
                output_folder = match.group(1)
            else:
                print("Output data folder not found in the output. Exiting.")
                return
        else:
            print(f"Error: {subprocess_result.stderr}")
            return  # exit the function if the subprocess failed

        # process the update and submit to Nema
        workflow = Workflow(
            global_id=existing_workflow.global_id,
            name=existing_workflow.name,
            description=existing_workflow.description,
            output_folder=output_folder,
        )

        workflow.process_update()

        print(
            f"Workflow '{this_identifier}' successfully run \U0001F680 and results uploaded to Nema. \U00002705"
        )


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
import os
import json
import requests
import mimetypes
from typing import List, Optional

from nema.utils.global_config import GLOBAL_CONFIG


def get_mime_type(file_path):
    mime_type, _ = mimetypes.guess_type(file_path)
    return mime_type or "application/octet-stream"


NEMA_PATH = os.path.expanduser("~/.nema")
AUTH_PATH = os.path.join(NEMA_PATH, "auth")


def save_auth_data(refresh_token: str, access_token: str):
    os.makedirs(os.path.dirname(AUTH_PATH), exist_ok=True)

    with open(AUTH_PATH, "w") as f:
        f.write(
            json.dumps({"refresh_token": refresh_token, "access_token": access_token})
        )


@dataclass
class ConnectivityManager:
    _auth_token: str = ""
    _refresh_token: str = ""
    _user_id: Optional[str] = None
    _is_auth_set: bool = False

    def __post_init__(self):

        if os.path.exists(AUTH_PATH):
            with open(AUTH_PATH, "r") as f:
                data = json.load(f)
                self._auth_token = data.get("access_token", "")
                self._refresh_token = data.get("refresh_token", "")

            if self._auth_token:
                self._is_auth_set = True
        else:
            self._is_auth_set = False

        if not self._is_auth_set:
            self._user_id = os.getenv("NEMA_USER_ID")

    @property
    def auth_token(self):
        return self._auth_token

    def get_latest_commit(self) -> str:

        data_api_url = (
            f"{GLOBAL_CONFIG.project_api_url}/artifacts/overview/commit-history"
        )

        response = requests.get(
            data_api_url,
            headers=self.get_headers(),
            params={"count": 1},
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.get_latest_commit()
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        data = response.json()

        return data["history"][0]["commit_hash"]

    def pull_blob_data(
        self,
        global_id: int,
        folder: str,
        commit_id: str,
        branch: str = "",
        requested_filename: Optional[str] = None,
    ):
        data_api_url = (
            f"{GLOBAL_CONFIG.project_api_url}/artifacts/data/{global_id}/download"
        ) + (f"/{requested_filename}" if requested_filename else "")

        response = requests.get(
            data_api_url,
            headers=self.get_headers(),
            stream=True,
            params={"commit_id": commit_id, "branch": branch},
        )

        if response.status_code == 413:
            # file is too large so we need to use a presigned link
            presigned_api_url = (
                f"{GLOBAL_CONFIG.project_api_url}/artifacts/data/{global_id}/download-link"
            ) + (f"/{requested_filename}" if requested_filename else "")

            response_presigned_api_url = requests.get(
                presigned_api_url,
                headers=self.get_headers(),
                stream=True,
                params={"commit_id": commit_id},
            )
            response_presigned_api_url.raise_for_status()
            data_response_presigned_api_url = response_presigned_api_url.json()
            link = data_response_presigned_api_url["link"]

            # Actually download the file
            response = requests.get(link, stream=True)

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.pull_blob_data(global_id, folder, commit_id, branch)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        # Raise an error if the request fails
        response.raise_for_status()

        # Try to get the filename from the 'Content-Disposition' header
        filename = None

        if "Content-Disposition" in response.headers:
            content_disposition = response.headers.get(
                "Content-Disposition", response.headers.get("content-disposition")
            )
            if "filename=" in content_disposition:
                filepath = content_disposition.split("filename=")[-1].strip('"')
                filename = os.path.basename(filepath).split("/")[-1]

        # If no filename in headers, infer from the URL
        if filename is None:
            filename = f"DATA_{global_id}"

        # Full path for saving the file
        fldr_path = os.path.join(folder, f"{global_id}")
        os.makedirs(fldr_path, exist_ok=True)
        filepath = os.path.join(fldr_path, filename)

        # Write the file in chunks to avoid memory overload
        with open(filepath, "wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

        return f"{global_id}/{filename}"

    def get_headers(self):
        if not self._auth_token:
            return {"X-User-Id": self._user_id}

        return {"Authorization": f"Bearer {self.auth_token}"}

    def refresh_auth_token(self):
        """Refresh the authentication token"""

        tenant_url = GLOBAL_CONFIG.tenant_api_url
        login_url = f"{tenant_url}/authentication/refresh-token"

        response = requests.post(login_url, json={"token": self._refresh_token})

        if not response.ok:
            if response.status_code == 401:
                raise ConnectionError("Token refresh failed, please login again.")
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        response_data = response.json()
        self._refresh_token = response_data["refresh_token"]
        self._auth_token = response_data["access_token"]

        save_auth_data(self._refresh_token, self._auth_token)

    def pull_in_memory_data(
        self, global_id: int, commit_id: str = "", branch: str = ""
    ):

        data_api_url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/data/{global_id}"

        response = requests.get(
            data_api_url,
            headers=self.get_headers(),
            params={"commit_id": commit_id, "branch": branch},
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.pull_in_memory_data(global_id, commit_id, branch)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        return response.json()

    def push_data_update(
        self,
        global_id: int,
        form_data: dict,
        files: Optional[list] = None,
        branch: Optional[str] = None,
    ):

        url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/data/{global_id}"

        if files is None:
            # add a dummy file to force multipart
            files_to_send = [("dummy_file", ("dummy", b"", "application/octet-stream"))]

        else:
            files_to_send = [
                ("file", (fp.split("/")[-1], open(fp, "rb"), get_mime_type(fp)))
                for fp in files
            ]

        response = requests.put(
            url,
            headers={**self.get_headers()},
            data=form_data,
            files=files_to_send,
            params={"branch": branch},
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.push_data_update(global_id, form_data, files, branch)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        response.raise_for_status()

        return response

    def push_workflow_update(
        self, global_id: int, workflow_raw_data: dict, files: List[str] = []
    ):

        url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/workflows/{global_id}/upload-external-results"

        files = [
            ("file", (fp.split("/")[-1], open(fp, "rb"), get_mime_type(fp)))
            for fp in files
        ]

        if (
            len(files) == 0
        ):  # until the API can support non-multipart, we need to do this, really dumb..
            files.append(("file", ("dummy", b"dummy", "application/octet-stream")))

        response = requests.post(
            url,
            headers={**self.get_headers()},
            data=workflow_raw_data,
            files=files,
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.push_workflow_update(global_id, workflow_raw_data)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        return response.json()

    def push_workflow_outputs(
        self,
        global_id: int,
        workflow_raw_data: dict,
        branch: Optional[str] = None,
        files: List[str] = [],
    ):

        url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/workflows/{global_id}/upload-internal-only-workflow-outputs"

        files = [
            ("file", (fp.split("/")[-1], open(fp, "rb"), get_mime_type(fp)))
            for fp in files
        ]

        if (
            len(files) == 0
        ):  # until the API can support non-multipart, we need to do this, really dumb..
            files.append(("file", ("dummy", b"dummy", "application/octet-stream")))

        response = requests.post(
            url,
            headers={**self.get_headers()},
            data=workflow_raw_data,
            files=files,
            params={"branch": branch},
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.push_workflow_update(global_id, workflow_raw_data)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        return response.json()

    def create_workflow(self, workflow_data: dict):
        """Create a new workflow in the nema platform"""

        url = f"{GLOBAL_CONFIG.project_api_url}/artifacts/workflows"

        response = requests.post(
            url,
            headers={**self.get_headers(), "Content-Type": "application/json"},
            json=workflow_data,
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.create_workflow(workflow_data)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        return_data = response.json()
        return return_data["created_global_id"]

    def retrieve_app_code(
        self,
        global_id: int,
        commit_id: Optional[str] = None,
        branch: Optional[str] = None,
    ):
        data_api_url = (
            f"{GLOBAL_CONFIG.project_api_url}/artifacts/apps/{global_id}/download"
        )

        response = requests.get(
            data_api_url,
            headers=self.get_headers(),
            stream=True,
            params={"commit_id": commit_id, "branch": branch},
        )

        if not response.ok:
            if response.status_code == 401:
                self.refresh_auth_token()
                return self.retrieve_app_code(global_id, commit_id, branch)
            else:
                raise ConnectionError(
                    f"Connection error: {response.status_code} -- {response.text}"
                )

        # Raise an error if the request fails
        response.raise_for_status()

        return response.text



from dataclasses import dataclass, field
import shutil
import tempfile
import os
import pint
from typing import Optional, List

from nema.utils.file_name import generate_random_file_name
from nema.utils.units import UNIT_REGISTRY


@dataclass
class DataProperties:

    @property
    def is_blob_data(self):
        return False

    @property
    def data_type(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    def get_value(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    def __nema_marshall__(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        raise NotImplementedError("This method should be implemented by the subclass")

    def close(self):
        pass

    def get_contents(self):
        raise NotImplementedError("This method should be implemented by the subclass")

    def get_file_name(self, *args, **kwargs):
        raise NotImplementedError("This method should be implemented by the subclass")

    def set_file_name(self, *args, **kwargs):
        raise NotImplementedError("This method should be implemented by the subclass")


@dataclass
class SingleValue(DataProperties):

    def __nema_marshall__(self):
        return {"value": self.value}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(value=data["value"])

    def get_value(self):
        return self.value


@dataclass
class BooleanValue(SingleValue):
    value: bool = False

    @property
    def data_type(self):
        return "BOOL"


@dataclass
class StringValue(SingleValue):
    value: str = ""

    @property
    def data_type(self):
        return "STRING"


@dataclass
class IntegerValue(SingleValue):
    value: int = 0

    @property
    def data_type(self):
        return "INT"


@dataclass
class FloatValue(SingleValue):
    value: float = 0.0

    @property
    def data_type(self):
        return "FLOAT"


@dataclass
class CurrencyValue(FloatValue):
    currency_code: str = ""

    def __nema_marshall__(self):
        return {"value": self.value, "currency_code": self.currency_code}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(value=data["value"], currency_code=data["currency_code"])

    @property
    def data_type(self):
        return "CURRENCY"


@dataclass
class FloatValueWithArbitraryUnit(SingleValue):
    value: float = 0.0
    units: str = ""

    def __nema_marshall__(self):
        return {"value": self.value, "units": self.units}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(value=data["value"], unit=data["units"])


@dataclass
class IntValueWithArbitraryUnit(FloatValueWithArbitraryUnit):
    value: int = 0

    @property
    def data_type(self):
        return "INT_WITH_PHYSICAL_UNIT.V0"


@dataclass
class FloatValueWithPhysicalUnit(SingleValue):

    value: pint.Quantity = 0.0

    @property
    def data_type(self):
        return "FLOAT_WITH_PHYSICAL_UNIT.V0"

    def __nema_marshall__(self):
        unit_str = format(self.value.units, "~").replace(" ", "").replace("**", "^")
        return {"value": self.value.magnitude, "units": unit_str}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        value = data["value"] * UNIT_REGISTRY(data["units"])
        return cls(value=value)


@dataclass
class IntValueWithPhysicalUnit(FloatValueWithPhysicalUnit):

    @property
    def data_type(self):
        return "INT_WITH_PHYSICAL_UNIT.V0"


@dataclass
class BlobDataProperties(DataProperties):

    @property
    def is_blob_data(self):
        return True


@dataclass
class FileDataProperties(BlobDataProperties):
    _file_name: str = ""
    _local_file_name: str = ""
    _input_folder: str = ""
    _temp_file_path: str = field(init=False, default=None)

    def __post_init__(self):
        file_extension = self.get_default_file_extension()
        suffix = f".{file_extension}" if file_extension else ""
        self._temp_file_path = tempfile.mkstemp(suffix=suffix)[1]

    def __nema_marshall__(self):
        return {"file_name": self._file_name}

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(_file_name=data["file_name"])

    def close(self):
        if os.path.exists(self._temp_file_path):
            os.remove(self._temp_file_path)

    def __del__(self):
        self.close()

    @staticmethod
    def get_default_file_extension():
        return None

    def get_file_name(self, *args, **kwargs):
        return self._local_file_name

    def set_file_name(self, file_name: str, *args, **kwargs):

        # extract file extension
        _, file_extension = os.path.splitext(file_name)

        self.close()  # remove path without file extension
        self._temp_file_path = tempfile.mkstemp(suffix=file_extension)[1]

        self._local_file_name = file_name

    @property
    def file_extension(self):
        return self._file_name.split(".")[-1]

    def get_file_name_to_save(self):
        return self._temp_file_path

    def write_data_to_file_and_return_file_name(self, destination_folder: str):

        # move to destination folder
        output_file_name = generate_random_file_name(self.file_extension)
        destination_file_path = os.path.join(destination_folder, output_file_name)
        shutil.move(self._temp_file_path, destination_file_path)

        return output_file_name


@dataclass
class ArbitraryFile(FileDataProperties):

    @property
    def data_type(self):
        return "ARBITRARY_FILE.V0"


@dataclass
class SingleFileInCollection:
    display_filename: str  # from the API
    local_filename: Optional[str] = None


@dataclass
class ArbitraryFileCollection(DataProperties):

    files: List[SingleFileInCollection] = field(default_factory=list)

    @property
    def data_type(self):
        return "ARBITRARY_FILE_COLLECTION.V0"

    @classmethod
    def __nema_unmarshall__(cls, data: dict):
        return cls(
            files=[
                SingleFileInCollection(display_filename=file["display_filename"])
                for file in data["files"]
            ]
        )

    def get_file_name(self, requested_filename: str):
        for file in self.files:
            if file.display_filename == requested_filename:
                return file.local_filename
        return None

    def set_file_name(self, local_filename: str, requested_filename: str):
        for file in self.files:
            if file.display_filename == requested_filename:
                file.local_filename = local_filename
                return
        raise ValueError(f"File with name {requested_filename} not found in collection")



from dataclasses import dataclass

from nema.data.data_properties import DataProperties, FileDataProperties


@dataclass
class FigureDataProperties(DataProperties):
    pass


@dataclass
class Image(FigureDataProperties, FileDataProperties):

    @staticmethod
    def get_default_file_extension():
        return "png"

    @property
    def data_type(self):
        return "IMAGE.V0"
